;; Generated package description from popper.el  -*- no-byte-compile: t -*-
(define-package "popper" "0.4.6.0.20230908.183054" "Summon and dismiss buffers as popups" '((emacs "26.1")) :commit "570b0820f884a9c0e3d9cb07e7f7f523b39b836f" :authors '(("Karthik Chikmagalur" . "karthik.chikmagalur@gmail.com")) :maintainer '("Karthik Chikmagalur" . "karthik.chikmagalur@gmail.com") :keywords '("convenience") :url "https://github.com/karthink/popper")
